#ifndef __ManualIntersector_h__
#define __ManualIntersector_h__

#include "File.h"
#include "Stringf.h"

class ManualIntersector
{
public:
    void intersect( ThreadStateGraph &threadOne, ThreadStateGraph &threadTwo, ThreadStateGraph &output )
    {
        storeGraphs( threadOne, threadTwo, output );
        initializeOutputGraph();
        createOutputStates();
        createOutputEdges();
    }

private:
    ThreadStateGraph *_threadOne;
    ThreadStateGraph *_threadTwo;
    ThreadStateGraph *_output;

    void storeGraphs( ThreadStateGraph &threadOne, ThreadStateGraph &threadTwo, ThreadStateGraph &output )
    {
        _threadOne = &threadOne;
        _threadTwo = &threadTwo;
        _output = &output;
    }

    void initializeOutputGraph()
    {
        _output->_events = _threadOne->_events;
        _output->_initialState = fuseLabels( _threadOne->_initialState, _threadTwo->_initialState );
    }

    void createOutputStates()
    {
        Map<String, ThreadState>::iterator iteratorOne;
        Map<String, ThreadState>::iterator iteratorTwo;

        for ( iteratorOne = _threadOne->_graph.begin(); iteratorOne != _threadOne->_graph.end(); ++iteratorOne )
        {
            for ( iteratorTwo = _threadTwo->_graph.begin(); iteratorTwo != _threadTwo->_graph.end(); ++iteratorTwo )
            {
                addCombinedState( iteratorOne, iteratorTwo );
            }
        }
    }

    void addCombinedState( Map<String, ThreadState>::iterator &iteratorOne,
                           Map<String, ThreadState>::iterator &iteratorTwo )
    {
        String newLabel = fuseLabels( iteratorOne, iteratorTwo );

        ThreadState newState;
        newState._label = newLabel;

        fuseStates( iteratorOne->second, iteratorTwo->second, newState );

        _output->_graph[newLabel] = newState;
    }

    static String fuseLabels( Map<String, ThreadState>::iterator &iteratorOne,
                              Map<String, ThreadState>::iterator &iteratorTwo )
    {
        return fuseLabels( iteratorOne->first, iteratorTwo->first );
    }

    static String fuseLabels( const String &labelOne, const String &labelTwo )
    {
        return Stringf( "%s %s", labelOne.ascii(), labelTwo.ascii() );
    }

    void fuseStates( ThreadState &first, ThreadState &second, ThreadState &fused )
    {
        fused._requestedEvents = first._requestedEvents + second._requestedEvents;
        fused._watchedEvents = first._watchedEvents + second._watchedEvents;
        fused._blockedEvents = first._blockedEvents + second._blockedEvents;

        fused._requestedEvents.uniqueElements();
        fused._watchedEvents.uniqueElements();
        fused._blockedEvents.uniqueElements();

        for ( unsigned i = 0; i < fused._blockedEvents.size(); ++i )
        {
            Event blockedEvent = fused._blockedEvents[i];

            if ( fused._watchedEvents.exists( blockedEvent ) )
            {
                fused._watchedEvents.eraseByValue( blockedEvent );
            }
            if ( fused._requestedEvents.exists( blockedEvent ) )
            {
                fused._requestedEvents.eraseByValue( fused._blockedEvents[i] );
            }
        }

        for ( unsigned i = 0; i < fused._requestedEvents.size(); ++i )
        {
            Event requestedEvent = fused._requestedEvents[i];

            if ( fused._watchedEvents.exists( requestedEvent ) )
            {
                fused._watchedEvents.eraseByValue( requestedEvent );
            }
        }

        for ( unsigned i = 0; i < first._atomicPropositions.size(); ++i )
        {
            fused._atomicPropositions.add( first._atomicPropositions.at( i ) );
        }

        for ( unsigned i = 0; i < second._atomicPropositions.size(); ++i )
        {
            fused._atomicPropositions.add( second._atomicPropositions.at( i ) );
        }
    }

    void createOutputEdges()
    {
        Map<String, ThreadState>::iterator iteratorOne;
        Map<String, ThreadState>::iterator iteratorTwo;

        for ( iteratorOne = _threadOne->_graph.begin(); iteratorOne != _threadOne->_graph.end(); ++iteratorOne )
        {
            for ( iteratorTwo = _threadTwo->_graph.begin(); iteratorTwo != _threadTwo->_graph.end(); ++iteratorTwo )
            {
                addEdgesOfSingleState( iteratorOne, iteratorTwo );
            }
        }
    }

    void addEdgesOfSingleState( Map<String, ThreadState>::iterator &iteratorOne,
                                Map<String, ThreadState>::iterator &iteratorTwo )
    {
        ThreadState &combined = _output->_graph[fuseLabels( iteratorOne, iteratorTwo )];

        for ( unsigned i = 0; i < combined._requestedEvents.size(); ++i )
        {
            String firstSuccessor = iteratorOne->second.labelOfSuccessor( combined._requestedEvents[i] );
            String secondSuccessor = iteratorTwo->second.labelOfSuccessor( combined._requestedEvents[i] );

            String fusedSuccessor = fuseLabels( firstSuccessor, secondSuccessor );

            Transition transition( combined._requestedEvents[i], fusedSuccessor );
            combined._successors.append( transition );

            Transition backwardTransition( combined._requestedEvents[i], combined._label );
            _output->_graph[fusedSuccessor]._predecessors.append( backwardTransition );
        }

        for ( unsigned i = 0; i < combined._watchedEvents.size(); ++i )
        {
            String firstSuccessor = iteratorOne->second.labelOfSuccessor( combined._watchedEvents[i] );
            String secondSuccessor = iteratorTwo->second.labelOfSuccessor( combined._watchedEvents[i] );

            String fusedSuccessor = fuseLabels( firstSuccessor, secondSuccessor );

            Transition transition( combined._watchedEvents[i], fusedSuccessor );
            combined._successors.append( transition );

            Transition backwardTransition( combined._watchedEvents[i], combined._label );
            _output->_graph[fusedSuccessor]._predecessors.append( backwardTransition );
        }

        for ( unsigned i = 0; i < _output->_events.size(); ++i )
        {
            Event event = _output->_events[i];

            if ( ( !combined._requestedEvents.exists( event ) ) &&
                 ( !combined._watchedEvents.exists( event ) ) &&
                 ( !combined._blockedEvents.exists( event ) ) )
            {
                Transition transition( event, combined._label );
                combined._successors.append( transition );
                combined._predecessors.append( transition );
            }
        }
    }
};

#endif // __ManualIntersector_h__

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
